from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand

# we use the version that is used in klustakwik2 rather than separately testing the numba/cython
# versions
from klustakwik2.clustering import (accumulate_cluster_mask_sum, compute_cluster_means)

from .test_compute_cluster_masks import generate_simple_test_kk
from .test_compute_covariance_matrix import test_compute_covariance_matrix

def test_m_step():
    '''
    This test doesn't add much because it's more or less the same as the M_step() code, but
    I guess it guards against future errors that might be introduced.
    '''
    kk = generate_simple_test_kk(points_for_cluster_mask=0.5)

    num_cluster_members = kk.num_cluster_members
    num_clusters = kk.num_clusters_alive
    num_features = kk.num_features

    kk.M_step()
    
    cov_matrices = test_compute_covariance_matrix()
    
    for cluster, cov in enumerate(cov_matrices):
        cluster += 1        
        assert_array_almost_equal(cov.block, kk.covariance[cluster].block)
        assert_array_almost_equal(cov.diagonal, kk.covariance[cluster].diagonal)
    
    
if __name__=='__main__':
    test_m_step()



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand
from copy import deepcopy
from six.moves import range

# we use the version that is used in klustakwik2 rather than separately testing the numba/cython
# versions
from klustakwik2.clustering import (accumulate_cluster_mask_sum, compute_cluster_means,
                                    compute_covariance_matrices)

from .test_compute_cluster_masks import generate_simple_test_kk

def test_compute_covariance_matrix():
    kk = generate_simple_test_kk(points_for_cluster_mask=0.5)

    num_cluster_members = kk.num_cluster_members
    num_clusters = kk.num_clusters_alive
    num_features = kk.num_features

    kk.compute_cluster_masks()
    kk.cluster_mean = compute_cluster_means(kk)

    cov_matrices = []

    compute_covariance_matrices(kk)

    for cluster in range(1, num_clusters):
        cov = kk.covariance[cluster]
        f2m = kk.orig_features-kk.cluster_mean[cluster, :][newaxis, :]
        ct = kk.orig_correction_terms
        spikes = kk.get_spikes_in_cluster(cluster)
        block = zeros_like(cov.block)
        unmasked = cov.unmasked
        for spike in spikes:
            if cluster==1:
                for i in range(len(unmasked)):
                    block[i, i] += f2m[spike, unmasked[i]]**2
            else:
                block[:, :] += f2m[spike, unmasked][:, newaxis]*f2m[spike, unmasked][newaxis, :]
            for i in range(len(unmasked)):
                block[i, i] += ct[spike, unmasked[i]]
        if cluster==1:
            point = kk.mua_point
        else:
            point = kk.noise_point
        for i in range(len(cov.unmasked)):
            block[i, i] += point*kk.data.noise_variance[cov.unmasked[i]]
        factor = 1.0/(num_cluster_members[cluster]+point-1)
        block *= factor
        assert_array_almost_equal(block, cov.block)
        cov_matrices.append(deepcopy(cov))

    return cov_matrices


if __name__=='__main__':
    test_compute_covariance_matrix()



#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True

import numpy
cimport numpy

from cython cimport integral, floating

cpdef doaccum(integral[:] clusters,
              integral[:] unmasked,
              integral[:] ustart,
              integral[:] uend,
              floating[:] masks,
              integral[:] vstart,
              integral[:] vend,
              floating[:, :] cluster_mask_sum,
              integral num_special_clusters,
             ):
    cdef integral p, c, num_unmasked
    for p in range(len(clusters)):
        c = clusters[p]
        if c<num_special_clusters:
            continue
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            cluster_mask_sum[c, unmasked[ustart[p]+i]] += masks[vstart[p]+i]



from .m_step_cy import *
from six.moves import range

__all__ = ['compute_cluster_means', 'compute_covariance_matrices']

def get_diagonal(x):
    '''
    Return a writeable view of the diagonal of x
    '''
    return x.reshape(-1)[::x.shape[0]+1]




def compute_cluster_means(kk):
    data = kk.data
    num_clusters = len(kk.num_cluster_members)
    num_features = kk.num_features

    cluster_mean = numpy.zeros((num_clusters, num_features))
    num_added = numpy.zeros((num_clusters, num_features), dtype=int)

    do_compute_cluster_means(
                kk.clusters, data.unmasked, data.unmasked_start, data.unmasked_end,
                data.features, data.values_start, data.values_end,
                cluster_mean, num_added,
                kk.num_cluster_members, data.noise_mean,
                kk.mua_point, kk.prior_point,
                kk.mua_cluster, kk.num_special_clusters)

    return cluster_mean


def compute_covariance_matrices(kk):
    data = kk.data
    num_cluster_members = kk.num_cluster_members
    num_clusters = len(num_cluster_members)
    num_features = kk.num_features

    for cluster in range(kk.first_gaussian_cluster, num_clusters):
        cov = kk.covariance[cluster]
        block = cov.block
        block_diagonal = get_diagonal(block)

        spikes_in_cluster = kk.spikes_in_cluster
        spikes_in_cluster_offset = kk.spikes_in_cluster_offset
        spike_indices = spikes_in_cluster[spikes_in_cluster_offset[cluster]:spikes_in_cluster_offset[cluster+1]]

        f2m = numpy.zeros(num_features)
        ct = numpy.zeros(num_features)

        if kk.use_mua_cluster and cluster==kk.mua_cluster:
            point = kk.mua_point
            do_var_accum_mua(spike_indices, kk.cluster_mean[cluster, :],
                             kk.data.noise_mean, kk.data.noise_variance,
                             cov.unmasked, block,
                             data.unmasked, data.unmasked_start, data.unmasked_end,
                             data.features, data.values_start, data.values_end,
                             f2m, ct, data.correction_terms, num_features,
                             )
        else:
            point = kk.prior_point
            do_var_accum(spike_indices, kk.cluster_mean[cluster, :],
                         kk.data.noise_mean, kk.data.noise_variance,
                         cov.unmasked, block,
                         data.unmasked, data.unmasked_start, data.unmasked_end,
                         data.features, data.values_start, data.values_end,
                         f2m, ct, data.correction_terms, num_features,
                         )

        # add correction term for diagonal
        cov.diagonal[:] += len(spike_indices)*data.noise_variance[cov.masked]

        # Add prior
        block_diagonal[:] += point*data.noise_variance[cov.unmasked]
        cov.diagonal[:] += point*data.noise_variance[cov.masked]

        # Normalise
        factor = 1.0/(num_cluster_members[cluster]+point-1)
        cov.block *= factor
        cov.diagonal *= factor
